/*
 * basicClient.hpp
 *
 *  Created on: 04/07/2016
 *      Author: rescue
 */

#ifndef CROSBOT_EXPLORE_BASIC_CLIENT_HPP_
#define CROSBOT_EXPLORE_BASIC_CLIENT_HPP_

#include <crosbot/geometry/poses.hpp>
#include <crosbot/handle.hpp>
#include <crosbot/controls/command.hpp>
#include <crosbot/controls/status.hpp>
#include <crosbot_explore/ExploreAction.h>

#include <actionlib/client/simple_action_client.h>
#include <geometry_msgs/PoseStamped.h>
#include <ros/publisher.h>
#include <ros/service.h>
#include <tf/transform_listener.h>

namespace crosbot_explore {

typedef actionlib::SimpleActionClient<crosbot_explore::ExploreAction> ExploreAClient;

class BasicExplorerClient : public crosbot::HandledObject {
public:
    BasicExplorerClient();
    virtual ~BasicExplorerClient();

    void configure();
    void startup();
    void shutdown();

    // Command callbacks
    void callback_navGoal(const geometry_msgs::PoseStampedConstPtr& targetPose);
    void callback_receivedCommand(const crosbot_msgs::ControlCommandPtr command);

    // Explore Action Server calls
    void actionserver_goalActive();
    void actionserver_goalReached(const actionlib::SimpleClientGoalState& state,
            const crosbot_explore::ExploreResultConstPtr& result);
    void actionserver_feedback(const crosbot_explore::ExploreFeedbackConstPtr& feedback);

private:
    // Configuration
    std::string aStarSrvName;
    std::string exploreAServerName;
    std::string navGoalSub_name;
    std::string robot_frame;
    std::string postrackResetPub_name;
    std::string slamResetPub_name;
    std::string world_frame;

    // Publishers/subscribers/services
    crosbot::CrosbotCommandPtr crosbotCommand;
    crosbot::CrosbotStatusPtr crosbotStatus;
    ros::ServiceClient aStarSrv;
    ros::Publisher pubPostrackReset;
    ros::Publisher pubGraphSlamReset;
    ros::Subscriber subNavGoal;
    tf::TransformListener tfListener;

    // Explorer action server client
    ExploreAClient *actionClient;

    /**
     * Set mode of crosbot_explore
     */
    void setExploreMode(int mode);

    /**
     * Get command points
     */
    crosbot::Pose3D getCommandPoint(const crosbot_msgs::ControlCommandPtr command);
    crosbot::Pose3D getCommandPose(const crosbot_msgs::ControlCommandPtr command);

    /**
     * Get A* path to origin and set path
     */
    void setAStarPose(crosbot::Pose3D targetPose, bool targetOrientation);
    void setAStarPose(const geometry_msgs::PoseStamped& targetPose, bool targetOrientation);

    void command_reset(const crosbot_msgs::ControlCommandPtr command);
};
typedef crosbot::Handle<BasicExplorerClient> BasicExplorerClientPtr;

} // namespace crosbot_explore

#endif /* CROSBOT_EXPLORE_BASIC_CLIENT_HPP_ */
