/*
 * explorerNode.hpp
 *
 *  Created on: 30/08/2016
 *      Author: rescue
 */

#ifndef CROSBOT_EXPLORE_EXPLORERNODE_HPP_
#define CROSBOT_EXPLORE_EXPLORERNODE_HPP_

#include <crosbot_explore/explorer.hpp>

#include <actionlib/server/simple_action_server.h>
#include <crosbot/thread.hpp>
#include <crosbot_explore/ExploreAction.h>
#include <crosbot_explore/GetPath.h>
#include <crosbot_explore/CrosbotExploreReconfigConfig.h>
#include <dynamic_reconfigure/server.h>
#include <nav_msgs/OccupancyGrid.h>
#include <nav_msgs/Path.h>
#include <ros/publisher.h>
#include <ros/service_server.h>
#include <ros/subscriber.h>
#include <tf/transform_listener.h>

#include <string>

namespace crosbot {

// Typedef for reconfigure
typedef dynamic_reconfigure::Server<crosbot_explore::CrosbotExploreReconfigConfig> ReconfigServer;

// Typedef for action server
typedef actionlib::SimpleActionServer<crosbot_explore::ExploreAction> ExploreAServer;

class ExplorerROSNode : public Explorer {
protected:
    std::string actionServer_name;
    std::string baseFrame;
    std::string gridSub_name;
    std::string historySub_name;
    std::string imagePub_name;
    std::string velPub_name;

    // Actionlib connections
    ExploreAServer *actionServer;
    int currentGoalId;

    // ROS Topics
    ros::Subscriber gridSub;
    ros::Subscriber historySub;
    ros::Publisher imagePub;
    ros::Publisher velPub;
    tf::TransformListener tfListener;

    ReadWriteLock rosLock;
    nav_msgs::OccupancyGridConstPtr latestMap;
    nav_msgs::PathConstPtr latestHistory;
    VoronoiGridPtr latestVoronoi;

public:
    ExplorerROSNode();
    virtual ~ExplorerROSNode();

    virtual void configure();
    virtual void shutdown();

    // Over-rides from crosbot::Explorer
    virtual VoronoiGridPtr getLatestVoronoi();
    virtual Pose getLatestPose();
    virtual Pose findDriveTarget(const VoronoiGrid& voronoi, const Pose& robot);
    virtual bool stopMotors();
    virtual void driveTo(const Pose& relativePosition);
    virtual void statusChanged(const std::string& status);

    // ROS callbacks
    void callbackOccGrid(const nav_msgs::OccupancyGridConstPtr& latestMap);
    void callbackHistory(const nav_msgs::PathConstPtr& latestHistory);

    // Reconfig callback
    void callback_reconfigure(crosbot_explore::CrosbotExploreReconfigConfig &config, uint32_t level);

    // Action Server callback
    void callback_explore(const crosbot_explore::ExploreGoalConstPtr& goal);

protected:
    /**
     * Subscribe to grid topic
     */
    void subscribeGrid(ros::NodeHandle& nh);

    /**
     * Get the latest pose of the robot, given the frame-of-reference and timestamp, rather than pulling it from the latestMap
     */
    virtual Pose getLatestPose(std::string frame_id, Time timestamp);

    /**
     * Define the necessary parameters for processing the goal mode.
     */
    virtual void handleExploreMode(const crosbot_explore::ExploreGoalConstPtr& goal,
            crosbot_explore::ExploreResult& result, bool& enableDrive);

    /**
     * Set a sequence of waypoints to follow, via a nav_msgs::Path
     * @return Were the waypoints could be configured
     */
    bool setWaypoints(nav_msgs::Path& navPath);

    /**
     * Publish Configuration to dynamic parameter server
     */
    void publishReconfig(VoronoiGrid::Constraints& constraints);

private:
    // Allow local voronoi constraints for this explorer (this allows the constraints to be reset between different modes)
    VoronoiGrid::Constraints localVoronoiConstraints;

    // Dynamic reconfiguration - only on config request
    ReconfigServer *reconfig_server;
    ReconfigServer::CallbackType reconfig_callbackType;
    boost::recursive_mutex mutexReconfig;
    bool gotReconfig;
};

} // namespace crosbot


#endif /* CROSBOT_EXPLORE_EXPLORERNODE_HPP_ */
