/*
 * explorerNode.hpp
 *
 *  Created on: 30/08/2016
 *      Author: rescue
 */

#ifndef CROSBOT_EXPLORE_MOVE_NODE_HPP_
#define CROSBOT_EXPLORE_MOVE_NODE_HPP_

#include <crosbot_explore/explorerROS.hpp>

#include <crosbot/handle.hpp>
#include <crosbot_explore/astar.hpp>
#include <crosbot_explore/astar/astarParameters.hpp>
#include <crosbot_explore/CrosbotAStarReconfigConfig.h>

namespace crosbot_explore {

class MoveMode : public ExploreMode {
public:
    enum {
        MODE_ASTAR  = 5
    };
};

} // namespace crosbot_explore

namespace crosbot {

// Typedef for A* specific reconfigure
typedef dynamic_reconfigure::Server<crosbot_explore::CrosbotAStarReconfigConfig> AStarReconfigServer;

// Forward reference
class MoveNode;

class PlanningThreadMove : public crosbot::Thread, public crosbot::HandledObject {
public:

    PlanningThreadMove(MoveNode *moveNode);
    virtual ~PlanningThreadMove();

    virtual void run();

private:
    MoveNode *moveNode;
};
typedef crosbot::Handle<PlanningThreadMove> PlanningThreadMovePtr;

class MoveNode : public ExplorerROSNode {
public:

    MoveNode();
    virtual ~MoveNode();

    virtual void configure();
    virtual void shutdown();

    // Planning loop for calculating updated A* path
    void planningLoop();

    // Over-rides from crosbot::Explorer
    // TODO: implement override
//    virtual void statusChanged(const std::string& status);

protected:
    // Reconfig callback
    void callback_astarOccGrid(const nav_msgs::OccupancyGridConstPtr& astarLatestMap);
    void callback_astarReconfigure(crosbot_explore::CrosbotAStarReconfigConfig &config, uint32_t level);

    // Astar Voronoi
    VoronoiGridPtr getLatestAstarVoronoi();

    // Over-rides from crosbot::Explorer
    virtual void handleExploreMode(const crosbot_explore::ExploreGoalConstPtr& goal,
            crosbot_explore::ExploreResult& result, bool& enableDrive);

    /**
     * Set a sequence of waypoints to follow, via an A* Plan.
     * Converts the plan to a nav_msgs::Path and calls ExplorerROSNode::setWaypoints(nav_msgs::Path)
     *  in the frame of navPathFrameId
     * @return Were the waypoints could be configured
     */
    bool setWaypoints(Plan& plan, const std::string& frameId, const std::string& navPathFrameId, const crosbot::Time time);

private:
    PlanningThreadMovePtr planningThread;

    // Configuration
    std::string astarImgPub_name;
    std::string astarGridSub_name;
    bool differentPlanningGrid;
    double replanRate;

    // ROS Publishers/Subscribers
    ros::Subscriber astarGridSub;

    // Dynamic reconfiguration - only on config request
    AStarReconfigServer *aStarReconfig_server;
    AStarReconfigServer::CallbackType aStarReconfig_callbackType;
    boost::recursive_mutex aStarMutexReconfig;

    // A* planner
    ReadWriteLock astarRosLock;
    AStarPlanner planner;
    ros::Publisher astarImagePub;
    nav_msgs::OccupancyGridConstPtr astarLatestMap;
    VoronoiGrid::Constraints astarVoronoiConstraints;
    VoronoiGridPtr astarLatestVoronoi;

    // Planning loop
    bool planningOperating;
    ReadWriteLock dataLock;
    Pose3DStamped currentAStarGoal;

    /**
     * Publish A* Configuration to dynamic parameter server
     */
    void publishReconfig(AStarParameters& params);

    /**
     * Store current A* goal, taking care of concurrency issues
     */
    void setCurrentAStarGoal(const Pose3DStamped& goal);

    /**
     * Get The A* goal, in the frame of reference of the voronoi map
     */
    Pose3DStamped getCurrentAStarGoal(std::string frame_id, Time timestamp);

    /**
     * Revise the sequence of waypoints for the robot to follow
     */
    bool updateWaypoints();
};

} // namespace crosbot


#endif /* CROSBOT_EXPLORE_MOVE_NODE_HPP_ */
