/*
 * module.hpp
 *
 *  Created on: 20/02/2013
 *      Author: mmcgill
 */

#ifndef CROSBOT_FASTSLAM_MODULE_HPP_
#define CROSBOT_FASTSLAM_MODULE_HPP_

#include <ros/ros.h>
#include "fastslam.hpp"

#include <sensor_msgs/LaserScan.h>
#include <tf/transform_listener.h>
#include <tf/transform_broadcaster.h>
#include <nav_msgs/Path.h>

namespace crosbot {

namespace fastslam {

#define DEFAULT_ODOMFRAME			"/icp"
#define DEFAULT_MAPFRAME			"/map"
#define DEFAULT_BASEFRAME			"/base_link"
#define DEFAULT_MAXWAIT4TRANSFORM	2.0			// [s]

class FastSLAMModule;
typedef Handle<FastSLAMModule> FastSLAMModulePtr;

class FastSLAMModule : public FastSLAMMap, virtual public MapListener {
protected:
	// ROS Configuration
    std::string baseFrame;
    std::string mapFrame;
	std::string odomFrame;
	std::string pub_historyName;
	std::string pub_mapName;
	std::string pub_particlesName;
	std::string sub_scanName;
	std::string sub_snapName;

	ros::Subscriber scanSub;
	ros::Subscriber snapSub;

	ros::Publisher gridPub;
	ros::Publisher historyPub;
	ros::Publisher particlesPub;

	tf::TransformListener tfListener;
    tf::TransformBroadcaster tfBroadcast;
public:

    FastSLAMModule(std::string name = "");
    virtual ~FastSLAMModule();

    void initialize(ros::NodeHandle& nh);
    void shutdown();

	// ROS Callbacks
    void callbackScan(const sensor_msgs::LaserScanConstPtr& latestScan);
	void callbackSnap(const crosbot_map::SnapMsgConstPtr& snap);

	void publishTransform(const Pose& robot, const Pose& odometry, const Time& time);

	// Map Listener callbacks
	void motionTracked(MapPtr map);
	void mapUpdated(MapPtr map);
	void tagAdded(MapPtr map, TagPtr tag);
	void tagChanged(MapPtr map, TagPtr tag);
};

} // namespace fastslam

} // namespace crosbot

#endif /* CROSBOT_FASTSLAM_MODULE_HPP_ */
