/*
 * parameters.hpp
 *
 *  FASTSlam Parameters
 *
 *  Created on: 28/04/2017
 *      Author: rescue
 */

#ifndef CROSBOT_FASTSLAM_PARAMETERS_HPP_
#define CROSBOT_FASTSLAM_PARAMETERS_HPP_

#include <crosbot/geometry/defines.hpp>
#include <crosbot/geometry/poses.hpp>
#include <crosbot_fastslam/constants.hpp>
#include <crosbot_fastslam/probtable.hpp>

#include <cmath>

namespace crosbot {
namespace fastslam {

struct FastSLAMParameters {
public:
    uint32_t numberOfParticles;
    double gain;
    float initialOccupiedOdds;
    double initialZ;

    double updateThresholdXY;
    double updateThresholdTheta;

    double minDistanceBetweenSnaps;

    crosbot::Pose searchPose;
    double searchDistance;
    double searchFOV;
    double maxSensorRange;

    unsigned int mapRows;
    unsigned int mapColumns;
    unsigned int patchRows;
    unsigned int patchColumns;

    double maxHeight;
    double minHeight;
    bool addMotionCopy;

    FastSLAMParameters() {
        numberOfParticles = DEFAULT_NUMPARTICLES;
        gain = DEFAULT_GAINVALUE;

        initialOccupiedOdds = ProbabilityTable.lodds(DEFAULT_UNOBSERVEDLODDS);
        initialZ = 0;

        updateThresholdXY = DEFAULT_DISTANCETHRESHOLD;
        updateThresholdTheta = DEG2RAD(DEFAULT_ANGULARTHRESHOLD);

        minDistanceBetweenSnaps = DEFAULT_SNAP_THRESHOLD;

        searchDistance = -1;
        searchFOV = DEG2RAD(60);

        mapRows = MULTIMAP_ROWS;
        mapColumns = MULTIMAP_COLUMNS;
        patchRows = HEIGHTMAP_ROWS;
        patchColumns = HEIGHTMAP_COLUMNS;

        minHeight = 0;
        maxHeight = INFINITY;
        maxSensorRange = INFINITY;
        addMotionCopy = false;
    }
};

} // namespace fastslam
} // namespace crosbot

#endif /* PARAMETERS_HPP_ */
