/*
 * occGridMerger.hpp
 *
 *  Merge laser scan into occupancy grid
 *
 *  Created on: 07/03/2017
 *      Author: timothyw
 */

#ifndef CROSBOT_MAP_OCCGRID_MERGER_HPP_
#define CROSBOT_MAP_OCCGRID_MERGER_HPP_

#include <ros/ros.h>

#include <crosbot/handle.hpp>
#include <crosbot/thread.hpp>
#include <nav_msgs/OccupancyGrid.h>
#include <sensor_msgs/LaserScan.h>
#include <tf/transform_listener.h>

namespace crosbot_map {

class OccGridMerger : public crosbot::HandledObject {
public:
    OccGridMerger();
    virtual ~OccGridMerger();

    /**
     * Clear data
     */
    void clear();

    /**
     * Update map
     */
    void updateMap(const nav_msgs::OccupancyGridConstPtr& map);

    /**
     *
     */
    void updateLaser(const sensor_msgs::LaserScanConstPtr scan);

    /**
     * Combine all maps
     */
    nav_msgs::OccupancyGridConstPtr generateMap();

protected:
    // Configuration
    double laserOccupiedProbability;
    double laserMinDist;

    // Store
    nav_msgs::OccupancyGridConstPtr map;
    sensor_msgs::LaserScanConstPtr scan;
    crosbot::PointCloudPtr debugCloud;

    // Concurrency controls
    crosbot::Mutex dataMutex;

    // Transform
    tf::TransformListener tfListener;
private:
};

typedef crosbot::Handle<OccGridMerger> OccGridMergerPtr;


} // namespace crosbot_map



#endif /* CROSBOT_MAP_OCCGRID_MERGER_HPP_ */
