/*
 * geotiff.hpp
 *
 *  Created on: 17/08/2016
 *      Author: rescue
 */

#ifndef CROSBOT_MAP_GEOTIFF_GEOTIFF_HPP_
#define CROSBOT_MAP_GEOTIFF_GEOTIFF_HPP_

#include <crosbot/handle.hpp>

#include <crosbot/thread.hpp>
#include <crosbot_map_geotiff/colour4f.hpp>
#include <crosbot_map_geotiff/geotiffConstants.hpp>
#include <nav_msgs/OccupancyGrid.h>
#include <nav_msgs/Path.h>
#include <ros/publisher.h>
#include <ros/subscriber.h>
#include <visualization_msgs/Marker.h>

#include <QImage>
#include <string>

namespace crosbot_map_geotiff {

class CrosbotGeotiff : public crosbot::HandledObject, crosbot::Thread {

public:
    CrosbotGeotiff();
    ~CrosbotGeotiff();

    void configure();
    void start();
    void stop();

    void run();

    // Callbacks
    void callback_markers(const visualization_msgs::MarkerConstPtr marker);
    void callback_map(const nav_msgs::OccupancyGridConstPtr map);
    void callback_path(const nav_msgs::PathConstPtr path);

private:
    bool operating;
    double frequency;

    // Configuration
    std::string map_topic;
    std::string mapImg_topic;
    std::string markers_topic;
    std::string path_topic;
    std::string tiff_filename;

    // ROS publishers/subscribers
    ros::Publisher mapImgPub;
    ros::Subscriber mapSub;
    ros::Subscriber markersSub;
    ros::Subscriber pathSub;

    // Saved content
    bool haveMap;
    std::map<int, visualization_msgs::Marker> markers;
    nav_msgs::OccupancyGridConstPtr map;
    nav_msgs::PathConstPtr history;

    // Thread management
    crosbot::Mutex processMutex;

    // Generates a Qt object for the geotiff
    QImage* produceQtGeotiff();
};
typedef crosbot::Handle<CrosbotGeotiff> CrosbotGeotiffPtr;

} // namespace crosbot_map_geotiff


#endif /* CROSBOT_MAP_GEOTIFF_GEOTIFF_HPP_ */
